/* CTA Version 1.2.2 by John Alroy

   date of release: 4.11.96
   requires CONJUNCT or NEXUS format input files
   previous releases:
    v. 1.2.1  2.11.96
    v. 1.2  1.10.96
    v. 1.1  30.9.96
    v. 1.0.4  30.4.96
    v. 1.0.3  16.4.96
    v. 1.0.2  18.10.95
    v. 1.0.1  4.8.95
    v. 1.0    1.2.95

   minor syntax fixes 4.11.96
   minor input fix 2.11.96
   prompts for character ordering if there is no DEFTYPE statement;
     understands tilde in SYMBOLS statement; outputs maximum connection
     "lengths"; outputs name of ancestral object to screen; bug fix in
     computation of binaries for missing data 1.10.96
   allows unordered characters; reads SYMBOLS statement; uses symbols if cell
     values are not numbers; bug fixes in output 30.9.96
   more input routine fixes; recognizes MISSING= and GAP= statements 25.4.96
   minor input routine debugging 16.4.96
   minor clean-up; TCI and TRI computed 16.10.95, 18.10.95
   daisy chains in polarized graphs disallowed 20.7.95
   complementary coding made optional 19.7.95
   polarized connections (14-16.1.95; unsuccessful attempt 14-18.9.94)
   determines input file type automatically; non-ANSI calls eliminated 14.9.94
   revised, CRR algorithm written 31.12.93-2.1.94
   converted to C 5.7.93, 6-8.9.93
 routines not yet written:
   cospeciation hypotheses treated as fragment mergers
   all connections through hypothetical objects - if so, convergence-on-
     PAUP warning when only n-1 connections are requested */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#define MAXCHARS 200
#define MAXATT 1200
#define MAXOBJ 250  /* DO NOT EXCEED 255 */
#define MAXREX 10000
#define MAXWIDTH 80
#define MAXTREE 1000

FILE *file1,*file2,*file3,*fopen();
int *cumst,*unscor,*mat,*group,*tgroup,*incircuit,*bot,*top;
int *temp,*temp2,*temp3,*temp4;
int *track,*otrack,*stone,*last,*treelen,*tancest;
char *str,*lstr,*rstr,*tstr,*symbstr,*tmat[MAXATT];
char *trees[MAXTREE],*trees2[MAXTREE],*nam[MAXOBJ];
char *link[MAXOBJ],*linkceil[MAXOBJ],*ancest[MAXOBJ];

char c;
int a,b,i,j,k,l,m,n,q,y,z,absval,tempint,tempint2,mintree;
int ii,jj,kk;
int nchar=0,tunscor=0,states=0,nobj=0,natt=0,nrex=0,maxfrags=0;
int oldl=0,ntrax,npresent=0,ntree=1,ontree=1;
int nunlink=0,nlink=0,onlink=0,ngroup,nrev,nbest=1,ntosearch=1;
int nincircuit=0,joint[99],nback=0,minback=0,maxdist=0;
int checking=1,temptrax,best,inputformat=0;
int constrained=0,maxlink,ret,hil,hir,grp1,grp2;
int polar=0,complement=0,ordered=-1;
int oldlinks[99][2];
long int searched=0;
float fx,hilink,expt,prob;

void NewRecord(int x)	{
  absval=abs(x);
  if (x==absval) temp2[x]++;
  if ((inputformat!=1)&(x>natt)) natt=x;
  states++;
  nrex++;
  mat[nrex]=x;
  if (inputformat==1) { track[nrex]=i; otrack[nrex]=i; }
  else { track[nrex]=nobj; otrack[nrex]=nobj; }
  if (x<0)	{
    track[nrex]=-track[nrex];
    otrack[nrex]=-otrack[nrex];
  }
  stone[nrex]=last[absval];
  last[absval]=nrex;
}

void BackTracks()	{
/* finds objects that fall in the circuit by eliminating each object
      and seeing if that breaks the circuit */
  int x;

  nincircuit=2;
  for (ii=3;ii<=nobj;ii++) incircuit[ii]=0;
  for (ii=1;ii<=nobj;ii++)
    if ((group[ii]==grp1)&(ii!=incircuit[1])&(ii!=incircuit[2]))	{
      for (jj=1;jj<=nobj;jj++) tgroup[jj]=jj;
      for (jj=1;jj<=onlink;jj++)
        if ((group[trees[ontree][jj]]==grp1)&(group[trees2[ontree][jj]]==grp1))
          if ((trees[ontree][jj]!=ii)&(trees2[ontree][jj]!=ii))	{
            grp2=tgroup[trees2[ontree][jj]];
            for (kk=1;kk<=nobj;kk++)
              if ((group[kk]==grp1)&(tgroup[kk]==grp2))	{
                tgroup[kk]=tgroup[trees[ontree][jj]];
              }
          }
      if (tgroup[incircuit[1]]!=tgroup[incircuit[2]])	{
 /* then circuit is broken */
        nincircuit++;
        incircuit[nincircuit]=ii; /* and therefore ii is in the circuit */
      }
    }
/* makes sure incircuit lists circuit members in correct order
    MAY NOT WORK IF MULTIPLE CIRCUITS OVERLAP */
  for (ii=1;ii<nincircuit;ii++)
    for (jj=ii+1;jj<=nincircuit;jj++)
if ((link[incircuit[ii]][incircuit[jj]]<-1)||(link[incircuit[jj]][incircuit[ii]]<-1))	{
          x=incircuit[ii+1]; incircuit[ii+1]=incircuit[jj];
          incircuit[jj]=x; jj=nincircuit;
        }
  nback=0; minback=9999;
 /* needed to make sure universal states do not count as back-tracks */
  for (ii=1;ii<=natt;ii++) { temp[ii]=0; temp2[ii]=0; }
  for (ii=1;ii<=nincircuit;ii++)
    for (jj=bot[incircuit[ii]];jj<=top[incircuit[ii]];jj++)
      temp[abs(mat[jj])]++;
  for (ii=1;ii<=natt;ii++)	{
    if (temp[ii]==nincircuit) temp[ii]=0; else temp[ii]=-1;
  }
 /* sees which objects have back-tracks, i.e., cannot be hybrids */
  for (ii=1;ii<=nincircuit;ii++)	{
    x=incircuit[ii-1]; y=incircuit[ii]; z=incircuit[ii+1];
    if (ii==1) x=incircuit[nincircuit];
    if (ii==nincircuit) z=incircuit[1];
    for (jj=1;jj<=natt;jj++) if (temp[jj]!=0) temp[jj]=-1;
    tempint=0;
    for (jj=bot[x];jj<=top[x];jj++)
      if ((temp[abs(mat[jj])]!=0)&(mat[jj]>0))
        temp[abs(mat[jj])]=x;
    for (jj=bot[y];jj<=top[y];jj++)
      if ((temp[abs(mat[jj])]==x)&(mat[jj]>0))
        temp[abs(mat[jj])]=y;
    for (jj=bot[z];jj<=top[z];jj++)
      if ((temp[abs(mat[jj])]==y)&(mat[jj]>0))	{
        tempint++;
        if (x>0) { nback++; x=0; }
      }
    if (x!=0) temp2[ii-nback]=incircuit[ii];
    if (tempint<minback) minback=tempint;
  }
}

void SwapConnections()	{

for (l=1;l<=maxlink;l++)
  for (m=1;m<=maxlink-l;m++)
    if (trees[ntree][m]>trees[ntree][m+1])	{
      tempint2=trees[ntree][m];
      trees[ntree][m]=trees[ntree][m+1];
      trees[ntree][m+1]=tempint2;
      tempint2=trees2[ntree][m];
      trees2[ntree][m]=trees2[ntree][m+1];
      trees2[ntree][m+1]=tempint2;
    }
    else if (trees[ntree][m]==trees[ntree][m+1])
      if (trees2[ntree][m]>trees2[ntree][m+1])	{
        tempint2=trees[ntree][m];
        trees[ntree][m]=trees[ntree][m+1];
        trees[ntree][m+1]=tempint2;
        tempint2=trees2[ntree][m];
        trees2[ntree][m]=trees2[ntree][m+1];
        trees2[ntree][m+1]=tempint2;
      }
}

void main()	{

int x;
char gapchar=45,misschar=63;

printf("\nCTA Version 1.2.2 by John Alroy\n");
printf("Date of release: 4.11.96\n");

cumst=(int*)calloc(MAXCHARS,sizeof(int));
unscor=(int*)calloc(MAXCHARS,sizeof(int));
mat=(int*)calloc(MAXREX,sizeof(int));
group=(int*)calloc(MAXOBJ,sizeof(int));
tgroup=(int*)calloc(MAXOBJ,sizeof(int));
incircuit=(int*)calloc(MAXOBJ,sizeof(int));
bot=(int*)calloc(MAXOBJ,sizeof(int));
top=(int*)calloc(MAXOBJ,sizeof(int));
temp=(int*)calloc(MAXATT,sizeof(int));
temp2=(int*)calloc(MAXATT,sizeof(int));
temp3=(int*)calloc(MAXATT,sizeof(int));
temp4=(int*)calloc(MAXATT,sizeof(int));
track=(int*)calloc(MAXREX,sizeof(int));
otrack=(int*)calloc(MAXREX,sizeof(int));
stone=(int*)calloc(MAXREX,sizeof(int));
last=(int*)calloc(MAXATT,sizeof(int));
treelen=(int*)calloc(MAXTREE,sizeof(int));
tancest=(int*)calloc(MAXOBJ,sizeof(int));
nam[0]=(char*)calloc(MAXWIDTH,sizeof(char));
str=(char*)calloc(MAXWIDTH,sizeof(char));
lstr=(char*)calloc(MAXWIDTH,sizeof(char));
rstr=(char*)calloc(MAXWIDTH,sizeof(char));
tstr=(char*)calloc(MAXWIDTH,sizeof(char));
symbstr=(char*)calloc(MAXWIDTH,sizeof(char));

printf("\nMatrix file: ");
fflush(stdout);	//++pac

scanf("%s",str);
file1=fopen(str,"r");
printf("Output file: ");
fflush(stdout);	//++pac
scanf("%s",str);
file2=fopen(str,"w");
printf("Complementary coding? "); fflush(stdout);	//++pac
scanf("%s",str);
if ((strncmp(str,"y",1)==0)||(strncmp(str,"Y",1)==0))
  complement=1;
printf("Polarize connections? ");fflush(stdout);	//++pac
 scanf("%s",str);
if ((strncmp(str,"y",1)==0)||(strncmp(str,"Y",1)==0))	{
  polar=1; tancest[1]=1;
}
else	{
  for (i=1;i<=MAXOBJ;i++) tancest[i]=1;
}
/* "CRR" = Connection Removal-Replacement algorithm */
printf("Use constraining connections? ");
fflush(stdout);	//++pac
scanf("%s",str);
if ((strncmp(str,"y",1)==0)||(strncmp(str,"Y",1)==0))	{
  printf("Constraint file: ");
  fflush(stdout);	//++pac
  scanf("%s",str);
  file3=fopen(str,"r"); constrained=1;
}

fscanf(file1,"%s",str);
if ((strcmp(str,"begin")==0)||(strcmp(str,"BEGIN")==0)) inputformat=1;
else if ((strncmp(str,"#nexus",6)==0)||(strncmp(str,"#NEXUS",6)==0)) inputformat=1;
if (inputformat==1)	{
/* inputs NEXUS format files */
  fscanf(file1,"%s",str);
  while ((strcmp(str,"matrix")!=0)&(strcmp(str,"MATRIX")!=0))	{
    fscanf(file1,"%s",str);
    if ((strncmp(str,"MISSING",7)==0)||(strncmp(str,"missing",7)==0))	{
      if (strchr(str,61)!=NULL) misschar=str[strlen(str)-1];
      else	{
        fscanf(file1,"%s",str); fscanf(file1,"%s",str); misschar=str[0];
      }
    }
    else if ((strncmp(str,"GAP",3)==0)||(strncmp(str,"gap",3)==0))	{
      if (strchr(str,61)!=NULL) gapchar=str[strlen(str)-1];
      else	{
        fscanf(file1,"%s",str); fscanf(file1,"%s",str); gapchar=str[0];
      }
    }
    else if ((strncmp(str,"SYMBOLS",7)==0)||(strncmp(str,"symbols",7)==0)) {
      if (strlen(str)>10) sscanf(str,"%8s%s",tstr,symbstr);
      else if (strlen(str)==7) fscanf(file1,"%s %s",str,symbstr);
      else printf("WARNING! Something is wrong with the SYMBOLS statement.\n");
      if (strchr(symbstr,126)!=NULL)	{
        strcpy(str,symbstr); y=0; z=0;
        for (i=48;i<=122;i++)	{
          if (i==59) i++;
          if (strchr(str,i)!=NULL) { y=i; i=122; }
        }
        for (i=122;i>=48;i--)	{
          if (i==59) i--;
          if (strchr(str,i)!=NULL) { z=i; i=48; }
        }
        if ((y>0)&(z>0))	{
          strcpy(symbstr,"");
          for (c=y;c<=z;c++) symbstr[c-y]=c;
          symbstr[z-y+1]=str[strlen(str)];
        }
        else printf("WARNING! Can't understand your SYMBOLS statement.\n");
      }
    }
    else if ((strncmp(str,"DEFTYPE",7)==0)||(strncmp(str,"deftype",7)==0)) {
      ordered=1;
      if (strlen(str)==7)	{
        fscanf(file1,"%s",str); fscanf(file1,"%s",str);
      }
      if (strstr(str,"UNORD")!=NULL) ordered=0;
      else if (strstr(str,"unord")!=NULL) ordered=0;
    }
  }
  fscanf(file1,"%s",str);
  while (strcmp(str,"[")==0)	{
    while (strcmp(str,"]")!=0) fscanf(file1,"%s",str);
    fscanf(file1,"%s",str);
  }

/* reads data matrix */
  while ((str[nchar]!=59)&(strcmp(nam[nobj],";")!=0))	{
    nobj++;
    nam[nobj]=(char*)calloc(MAXWIDTH,sizeof(char));
    tmat[nobj]=(char*)calloc(MAXCHARS,sizeof(char));
    if (nobj==1) strcpy(nam[1],str);
    else fscanf(file1,"%s",nam[nobj]);
    q=0;
    fscanf(file1,"%s",str);
    if (nobj==1) nchar=strlen(str);
    for (q=1;q<=nchar;q++)	{
      if ((str[q-1]!=gapchar)&(str[q-1]!=misschar))	{
/* tests to see if cell value is a number */
        if ((str[q-1]>47)&(str[q-1]<58)) tmat[nobj][q]=str[q-1]-48;
/* if not, checks cell value against SYMBOL string */
        else	{
          for (i=0;i<strlen(symbstr);i++)
            if (symbstr[i]==str[q-1]) { tmat[nobj][q]=i; i=strlen(symbstr); }
        }
      }
      else if (str[q-1]==gapchar)	{
        tmat[nobj][q]=-7; unscor[q]=1;
      }
      else tmat[nobj][q]=-1;
      if (tmat[nobj][q]>cumst[q]) cumst[q]=tmat[nobj][q];
    }
  }
  if (strcmp(nam[nobj],";")==0) nobj--;

  while (!feof(file1))	{
    fscanf(file1,"%s",str);	
    if ((strncmp(str,"DEFTYPE",7)==0)||(strncmp(str,"deftype",7)==0)) {
      ordered=1;
      if (strlen(str)==7)	{
        fscanf(file1,"%s",str); fscanf(file1,"%s",str);
      }
      if (strstr(str,"UNORD")!=NULL) ordered=0;
      else if (strstr(str,"unord")!=NULL) ordered=0;
    }
  }
  if (ordered==-1)	{
    printf("Characters are ordered (1) or unordered (2)? ");
    fflush(stdout);	//++pac
    scanf("%d",&ordered); if (ordered!=1) ordered=0;
  }

/* renumbers cell values so that the lowest observed value is zero */
  for (i=1;i<=nchar;i++)	{
    z=0; q=0;
    while (z==0)	{
      for (j=1;j<=nobj;j++)
        if (tmat[j][i]==q)	{
          z=1; j=nobj;
        }
      if (z==0) q++;
    }
    if (q>0)	{
      cumst[i]=cumst[i]-q;
      for (j=1;j<=nobj;j++) tmat[j][i]=tmat[j][i]-q;
    }
  }

/* for unordered characters, renumbers cell values to eliminate
     non-existent intermediate numbers */
  if (ordered==0)	{
    for (i=1;i<=nchar;i++)	{
      for (j=0;j<=cumst[i];j++)	{
        z=0;
        for (k=1;k<=nobj;k++)
          if (tmat[k][i]==j)	{
            z=1; k=nobj;
          }
        if (z==0)	{
          for (k=1;k<=nobj;k++)
            if (tmat[k][i]>j) tmat[k][i]--;
          cumst[i]--; j--;
        }
      }
    }
  }

/* computes maximal number of fragments for ordered characters, which
     requires finding the state for each character that is the closest
     to all others on average */
  for (i=1;i<=nchar;i++)	{
    if (ordered==1)	{
      z=9999;
      for (j=0;j<=cumst[i];j++)	{
        y=0;
        for (k=1;k<=nobj;k++)	{
          if (tmat[k][i]>=0) y=y+abs(tmat[k][i]-j);
        }
        if (y<z) { z=y; q=j; }
      }
      for (j=1;j<=nobj;j++)	{
        if (tmat[j][i]>=0) maxfrags=maxfrags+abs(tmat[j][i]-q);
      }
    }
/* for unordered states, the maximum is just the number of objects
     because CTA allows star graphs */ 
    else	{
      if (cumst[i]>0) maxfrags=maxfrags+nobj;
    }
  }

/* computes cumulative number of states */
  if ((complement==1)&(ordered==0))	{
    for (i=1;i<=nchar;i++)	{
      if (cumst[i]>0) cumst[i]=cumst[i]+cumst[i-1]+1;
      else cumst[i]=cumst[i-1];
    }
  }
  else
    for (i=1;i<=nchar;i++) cumst[i]=cumst[i]+cumst[i-1];
  natt=cumst[nchar];

/* computes extra ID numbers for characters with gaps */
  if ((complement==1)&(ordered==1)) tunscor=2*natt;
  else tunscor=natt;
  unscor[0]=tunscor;
  for (i=1;i<=nchar;i++)	{
    if (unscor[i]==1)	{
      tunscor=tunscor+cumst[i]-cumst[i-1]+1;
      unscor[i]=tunscor;
    }
    else unscor[i]=unscor[i-1];
  }

/* computes CONJUNCT-format binary codes, including complementary
     binaries */
  for (i=1;i<=nobj;i++)	{
    states=0;
    for (j=1;j<=nchar;j++)	{
      z=cumst[j-1];
      if (tmat[i][j]>-1)	{
  /* binaries for regular ordered cell values */
        if (ordered==1)	{
          for (k=z+1;k<=tmat[i][j]+z;k++)	{
            x=k; NewRecord(x);
          }
          if (complement==1)
            for (k=tmat[i][j]+z+1;k<=cumst[j];k++)	{
              x=k+natt; NewRecord(x);
            }
        }
  /* binaries for regular unordered cell values */
        else	{
          if ((tmat[i][j]>0)||(complement==1))	{
            if (cumst[j]>cumst[j-1]) { x=tmat[i][j]+z+1; NewRecord(x); }
          }
        }
      }
  /* binaries for missing or unscoreable cell values */
      else if (tmat[i][j]==-1)	{
        if (ordered==1)	{
          for (k=z+1;k<=cumst[j];k++)	{
            x=k*(-1); NewRecord(x);
          }
          if (complement==1)
            for (k=z+1;k<=cumst[j];k++)	{
              x=(k+natt)*(-1); NewRecord(x);
            }
          if (unscor[j]>unscor[j-1])
            for (k=unscor[j-1]+1;k<=unscor[j];k++)	{
              x=k*(-1); NewRecord(x);
            }
        }
  /* binaries for regular unordered cell values */
        else	{
          if ((tmat[i][j]>0)||(complement==1))
            if (cumst[j]>cumst[j-1])
              for (k=z+1;k<=cumst[j];k++)	{
                x=k*(-1); NewRecord(x);
              }
        }
      }
  /* binaries for gaps */
      else if (tmat[i][j]==-7)
        for (k=unscor[j-1]+1;k<=unscor[j];k++)	{
          x=k; NewRecord(x);
        }
    }
    bot[i]=nrex-states+1; top[i]=nrex;
  }
  if ((complement==1)&(ordered==1)) natt=natt*2;
}
else	{
/* input CONJUNCT format files */
  while (!feof(file1))	{
    nobj++; states=0;
    nam[nobj]=(char*)calloc(MAXWIDTH,sizeof(char));
    if (nobj>1) fscanf(file1,"%s\n",nam[nobj]);
    else strcpy(nam[nobj],str);
    fscanf(file1,"%s",str);
    while (strcmp(str,".")!=0)	{
      x=atoi(str); NewRecord(x);
      fscanf(file1,"%s",str);
    }
    fscanf(file1,"\n");
    bot[nobj]=nrex-states+1; top[nobj]=nrex;
  }
}
fclose(file1);
printf("Maximum number of connections (>=%d): ",nobj-1);
fflush(stdout);	//++pac
scanf("%d",&maxlink);
printf("\n");

/* initializes needed memory */
trees[1]=(char*)calloc(maxlink+1,sizeof(char));
trees2[1]=(char*)calloc(maxlink+1,sizeof(char));
for (i=1;i<=nobj;i++) link[i]=(char*)calloc(nobj,sizeof(char));
for (i=1;i<=nobj;i++) linkceil[i]=(char*)calloc(nobj,sizeof(char));
for (i=1;i<=nobj;i++) ancest[i]=(char*)calloc(nobj,sizeof(char));
for (i=1;i<=nobj;i++) ancest[i][i]=1;

/* reads constraint file */
if (constrained>0)	{
  fgets(str,80,file3);
  while ((!feof(file3))&(oldl<maxlink))	{
    fscanf(file3,"%s",lstr);
    fscanf(file3,"%s\n",rstr);
    fgets(str,80,file3);
    oldl++;
    for (i=1;i<=nobj;i++)
      if (strncmp(lstr,nam[i],20)==0) { oldlinks[oldl][1]=i; i=nobj; }
    for (i=1;i<=nobj;i++)
      if (strncmp(rstr,nam[i],20)==0) { oldlinks[oldl][2]=i; i=nobj; }
    if ((oldlinks[oldl][1]==0)||(oldlinks[oldl][2]==0))
      printf("Names do not match.\n");
  }
}

/* eliminates constant characters; AUTAPOMORPHIES MUST BE RETAINED */
ntrax=nrex;
for (i=1;i<=nrex;i++)	{
  if ((temp2[abs(mat[i])]==0)||(temp2[abs(mat[i])]==nobj))	{
    ntrax--; mat[i]=0; track[i]=0; otrack[i]=0;
  }
  else if (mat[i]<0)
    ntrax--;
}
for (i=1;i<=natt;i++)	{
  if (temp2[i]>0) npresent++;
  temp[i]=0; temp2[i]=0;
}

printf("%d objects\n",nobj);
if (polar!=0) printf("%s is the ancestral object\n",nam[1]);
if (nchar>0) printf("%d characters\n",nchar);
printf("%d attributes, %d informative\n",natt,npresent);
printf("%d fragments without any connections\n",ntrax);
if (oldl>0) printf("%d constraining connections\n",oldl);

ngroup=nobj;
for (i=1;i<=nobj;i++) group[i]=i;

/* computes potential links */
for (i=1;i<nobj;i++)	{
  if (top[i]-bot[i]+1>maxdist) maxdist=top[i]-bot[i]+1;
  for (j=0;j<=natt;j++) temp[j]=0;
  for (j=bot[i];j<=top[i];j++)
    temp[abs(mat[j])]=mat[j];
  for (j=i+1;j<=nobj;j++)	{
    for (k=bot[j];k<=top[j];k++)	{
 /* states are shared only if definitely present in BOTH objects */
      if ((mat[k]>0)&(temp[abs(mat[k])]>0))	{
        link[i][j]++; link[j][i]++;
      }
      if ((mat[k]!=0)&(temp[abs(mat[k])]!=0))	{
        linkceil[i][j]++; linkceil[j][i]++;
      }
    }
    if (link[i][j]>0)	{
      nunlink++;
    }
  }
}
printf("%d potential connections\n",nunlink);

/* computes initial graph using Greedy CTA algorithm */
hilink=9999;
while (nlink<maxlink)	{
  hilink=0; ret=0;
  if (nlink<oldl)	{
    hil=oldlinks[nlink+1][1];
    hir=oldlinks[nlink+1][2];
    hilink=link[hil][hir];
  }
  else
/*    for (i=1;i<nobj;i++)
      for (j=i+1;j<=nobj;j++) */
    for (i=1;i<=nobj;i++)
      for (j=1;j<=nobj;j++)
/* find highest link; if polarized, propagate network from root and
     disallow links that create daisy chains */
	if ((link[i][j]>hilink)&(tancest[i]>0)&(ancest[j][i]==0))	{
          minback=0;
/* DISCOUNTS BACKTRACKS
          if (group[i]==group[j])	{
            trees[1][nlink+1]=i; trees2[1][nlink+1]=j;
            grp1=group[i]; grp2=group[j];
            incircuit[1]=i; incircuit[2]=j; onlink=nlink;
            BackTracks();
          } */
          if (link[i][j]/*-minback*/>hilink)	{
	    hilink=link[i][j];
	    hil=i; hir=j;
          }
        }
  trees[1][nlink+1]=hil; trees2[1][nlink+1]=hir;
  grp1=group[hil]; grp2=group[hir];
  tancest[hir]++; ancest[hil][hir]=1;
  for (i=1;i<=nobj;i++)
    if (ancest[i][hil]==1)
      for (j=1;j<=nobj;j++)
        if (ancest[hir][j]==1) ancest[i][j]=1;

/* checks to see if isolated networks are joined */
  if (grp1!=grp2)	{
    link[hil][hir]=-7; link[hir][hil]=-7;
    ngroup--;
    for (i=1;i<=nobj;i++)
      if (group[i]==grp2)
        group[i]=grp1;
  }
  else {
    link[hil][hir]=-13; link[hir][hil]=-13;
    ret=1;
  }
/* finds triangles */
  for (i=1;i<=nobj;i++)	{
    x=hil; y=i;
   /* if (x>y)	{ tempint=x; x=y; y=tempint; } */
    if ((link[x][y]==-7)||(link[y][x]==-7))	{
      a=hir; b=i;
   /*   if (a>b) { tempint=a; a=b; b=tempint; } */
      if ((link[a][b]==-7)||(link[b][a]==-7))	{
	ret=3;
	i=nobj;
      }
    }
  }
  for (i=bot[hil];i<=top[hil];i++)	{
     temp[abs(mat[i])]=hil;
     temp2[abs(mat[i])]=i;
  }
  temp[0]=0; temp2[0]=0;

  for (i=bot[hir];i<=top[hir];i++)
    if (temp[abs(mat[i])]==hil)
      if (track[i]!=track[temp2[abs(mat[i])]])	{
        grp1=track[temp2[abs(mat[i])]];
        grp2=track[i];
        if (((grp1>0)&(grp2>0))&((grp2==hir)||(polar==0))) ntrax--;
        z=last[abs(mat[i])];
        while (z>0)	{
          if (track[z]==grp1)	{
            q=last[abs(mat[i])];
	    while (q>0)	{
	      if (track[q]==grp2)
                if ((mat[z]>0)&(mat[q]>0))	{
                  x=abs(otrack[z]);
	  	  y=abs(otrack[q]);
                  if ((polar==0)||(ancest[hil][y]>0))	{
                    if (link[x][y]>0) link[x][y]--;
                    if (link[x][y]==0)	{
                      nunlink--; link[x][y]=-1;
                    }
                    if (link[y][x]>0) link[y][x]--;
		    if (link[y][x]==0)	{
                      nunlink--; link[y][x]=-1;
                    }
		  }
	        }
	      q=stone[q];
	     }
	    }
          z=stone[z];
        }
 /* allows objects with unknown states to belong to conflicting tracks! */
        if (grp1<grp2)	{
          if (polar==0)	{
            tempint=grp1; grp1=grp2; grp2=tempint;
          }
          else if ((grp1<0)&(grp2>0))	{
            grp1=abs(grp1);
            z=last[abs(mat[i])];
            while (z>0)	{
	      if (track[z]==-grp1) track[z]=grp1;
              z=stone[z];
            }
          }
        }
        z=last[abs(mat[i])];
        while (z>0)	{
	  if (track[z]==grp2)
            if ((polar==0)||(ancest[hir][abs(otrack[z])]>0)) track[z]=grp1;
          z=stone[z];
        }
      }
  mat[0]=0;
  nunlink--; nlink++;
}

if ((complement==1)&(ordered==1))
  printf("%d tracks in initial graph\n\n",ntrax-(natt/2));
else
  printf("%d tracks in initial graph\n\n",ntrax);
printf("Search progress:\n Graphs   Best   Graphs   Not\n");
printf("Searched  Found   Found Swapped\n");
SwapConnections();

/* Connection Removal-Replacement algorithm */

treelen[1]=ntrax; mintree=ntrax;
while (checking<=ntree)	{

ntosearch--;

/* eliminates each link in turn and searches for a better replacement */
for (i=1;i<=maxlink;i++)	{
  hil=trees[checking][i]; hir=trees2[checking][i];

/* clears ancestry matrix */
  for (j=1;j<=nobj;j++)
    for (k=1;k<=nobj;k++) ancest[j][k]=0;
  for (j=1;j<=nobj;j++) ancest[j][j]=1;

/* colors matrix omitting the link that is being checked */
  temptrax=0;
  /* restores original matrix */
  for (j=1;j<=nrex;j++) { track[j]=otrack[j]; if (mat[j]>0) temptrax++; }
  for (j=1;j<=maxlink;j++)
    if (j!=i)	{
/* recomputes ancestry matrix */
      for (k=1;k<=nobj;k++)
        if (ancest[k][trees[checking][j]]==1)
          for (l=1;l<=nobj;l++)
            if (ancest[trees2[checking][j]][l]==1) ancest[k][l]=1;
      x=trees[checking][j]; y=trees2[checking][j];
      for (k=0;k<=natt;k++) temp[k]=0;
      for (k=bot[x];k<=top[x];k++)
        temp[abs(mat[k])]=track[k];
      for (k=bot[y];k<=top[y];k++)
        if ((track[k]!=0)&(temp[abs(mat[k])]!=0))
          if (track[k]!=temp[abs(mat[k])])	{
            grp1=temp[abs(mat[k])];
            grp2=track[k];
            if (((grp1>0)&(grp2>0))&((grp2==y)||(polar==0))) temptrax--;
            if (grp1<grp2)	{
              if (polar==0)	{
                tempint=grp1; grp1=grp2; grp2=tempint;
              }
              else if ((grp1<0)&(grp2>0))	{
                grp1=abs(grp1);
                z=last[abs(mat[k])];
                while (z>0)	{
	          if (track[z]==-grp1) track[z]=grp1;
                  z=stone[z];
                }
              }
            }
            z=last[abs(mat[k])];
            while (z>0)	{
              if (track[z]==grp2)
               if ((polar==0)||(ancest[y][abs(otrack[z])]>0)) track[z]=grp1;
              z=stone[z];
            }
          }
    }

/* recomputes link matrix */
  for (j=1;j<=nobj;j++)
    for (k=1;k<=nobj;k++)
      link[j][k]=0;
  for (j=1;j<=nobj;j++)	{
    for (k=0;k<=natt;k++) temp[k]=0;
    for (k=bot[j];k<=top[j];k++)
      temp[abs(mat[k])]=k;
    for (k=1;k<=nobj;k++)	{
      if ((k<=j)&(polar==0)) k=j+1;
      if ((temptrax-linkceil[j][k]<=mintree)&(j!=k))
        for (l=bot[k];l<=top[k];l++)
          if ((track[l]>0)&(track[temp[abs(mat[l])]]>0))
            if ((mat[l]!=0)&(mat[temp[abs(mat[l])]]!=0))
              if (track[l]!=track[temp[abs(mat[l])]])
                if ((polar==0)||(track[l]==k))
                  link[j][k]++;
    }
  }
  for (j=1;j<=maxlink;j++)	{
    link[trees[checking][j]][trees2[checking][j]]=-7;
    link[trees2[checking][j]][trees[checking][j]]=-7;
  }

/* finds strongest current link, disallowing daisy chains */
  best=0;
  for (j=1;j<=nobj;j++)
    for (k=1;k<=nobj;k++)
      if ((link[j][k]>best)&(ancest[k][j]==0))
        if ((j!=hil)||(k!=hir))
          best=link[j][k];
  searched=searched+(((nobj*nobj)-nobj)/2)-nobj+1;

/* creates a new tree if a better one has been found */
  if (temptrax-best<=mintree)	{
    for (j=1;j<=nobj;j++)	{
      for (k=1;k<=nobj;k++)	{
        if ((link[j][k]==best)&(ancest[k][j]==0))
          if ((j!=hil)||(k!=hir))	{
            for (l=1;l<=maxlink;l++)	{
              temp[l]=trees[checking][l];
              temp2[l]=trees2[checking][l];
            }
            temp[i]=j; temp2[i]=k;
/* sorting routine, needed to compare trees below */
            for (l=1;l<maxlink;l++)	{
              if (temp[l]>temp[l+1])	{
                tempint2=temp[l]; temp[l]=temp[l+1];
                temp[l+1]=tempint2;
                tempint2=temp2[l]; temp2[l]=temp2[l+1];
                temp2[l+1]=tempint2; l=l-2; if (l<0) l=0;
              }
              else if ((temp[l]==temp[l+1])&(temp2[l]>temp2[l+1]))	{
                tempint2=temp[l]; temp[l]=temp[l+1];
                temp[l+1]=tempint2;
                tempint2=temp2[l]; temp2[l]=temp2[l+1];
                temp2[l+1]=tempint2; l=l-2; if (l<0) l=0;
              }
            }
   /* confirms that tree is new */
            tempint=1;
            for (l=1;l<=ntree;l++)	{
              tempint=0;
              for (m=1;m<=maxlink;m++)
                if ((temp[m]!=trees[l][m])||(temp2[m]!=trees2[l][m]))	{
                  tempint=1; m=maxlink;
                }
              if (tempint==0) l=ntree; 
            }
            if (tempint==1)	{
              ntree++; treelen[ntree]=temptrax-link[j][k];
              if (treelen[ntree]<mintree)	{
                mintree=treelen[ntree]; nbest=1; ntosearch=1;
              }
              else	{
                nbest++; ntosearch++;
              }
              trees[ntree]=(char*)calloc(maxlink+1,sizeof(char));
              trees2[ntree]=(char*)calloc(maxlink+1,sizeof(char));
              for (l=1;l<=maxlink;l++)	{
                trees[ntree][l]=temp[l];
                trees2[ntree][l]=temp2[l];
              }
              SwapConnections();
            }
          }
      }
    }
  }
}
checking++;
while (treelen[checking]>mintree)
  checking++;
if (treelen[checking]==0) treelen[checking]=mintree;
printf("%7d  ",searched);
if ((complement==1)&(ordered==1)) printf("%5d   ",treelen[checking]-(natt/2));
else printf("%5d   ",treelen[checking]);
printf("%5d  %5d\n",nbest,ntosearch);
}

printf("\n"); fx=npresent; i=treelen[checking];
if ((complement==1)&(ordered==1)) { fx=fx-(natt/2); i=i-(natt/2); }
printf("TCI = %.0f/%d = %.3f\n",fx,i,fx/i);
printf("TRI = (%d-%d)/(%d-%.0f)",maxfrags,i,maxfrags,fx);
printf(" = %.3f\n",(maxfrags-i)/(maxfrags-fx));
printf("\n");
x=0;
for (i=1;i<=ntree;i++)	{
  if (treelen[i]==mintree)	{
    x++;
    fprintf(file2,"Tree %d (",x);
    if (polar==0) fprintf(file2,"un");
    fprintf(file2,"polarized):\n");
    for (j=1;j<=maxlink;j++)	{
      fprintf(file2,"%20.20s %20.20s  ",nam[trees[i][j]],nam[trees2[i][j]]);
      fprintf(file2,"%4d\n",maxdist-linkceil[trees[i][j]][trees2[i][j]]);
    }
    fprintf(file2,"\n");

    for (j=1;j<=nobj;j++) group[j]=j;
    for (j=1;j<=maxlink;j++)	{
      if (group[trees2[i][j]]!=group[trees[i][j]])	{
        tempint=group[trees2[i][j]];
        for (k=1;k<=nobj;k++)
          if (group[k]==tempint)
            group[k]=group[trees[i][j]];
      }
      else	{
        ret=1; ontree=i; onlink=j-1;
        grp1=group[trees[i][j]]; grp2=group[trees2[i][j]];
        incircuit[1]=trees[i][j]; incircuit[2]=trees2[i][j];
        BackTracks();
      }
    }
  }
}

}
